/* Lookup table for words. Currently, this just counts the number of
	unique words inserted into the table. Several different implementations
	of a table are supported: binary tree, binary splay tree, lexical
	search tree, and hash table. The specific implementation is selected
	at run-time using an integer selector; this is not as simple or
	elegant as OOP, but with plain C it will have to do.
	
	94/01/11 aih - added efficient allocation of strings instead of malloc
	94/01/06 aih - modifying allocation of items
	93/12/19 aih - created by splitting out of WordsInFile.c */

#include <stdio.h>
#include <string.h>
#include "LookupTableLib.h"
#include "MemoryLib.h"
#include "PointerFixedLib.h"
#include "ResourceConstantsLib.h"
#include "ResourceLib.h"

/* data stored in a table entry */
typedef struct {
	long count;
	char *key;
} LkpEntryType;
	
/* data stored in a tree's node */
typedef struct {
	NodeType node;
	LkpEntryType entry;
} NodeItemType;

/* data stored in hash tree entry */
typedef struct {
	LLType next;
	LkpEntryType entry;
} HashItemType;

/* data stored in a lexical tree node */
typedef struct {
	long count;
	CStr255 key;
} LexNodeItemType;

/*---------------------------------------------------------------------------*/
/* string allocation */
/*---------------------------------------------------------------------------*/

#define PTRSZ (8192)

/* header for a string pointer list */
typedef struct {
	LLType next;	/* next pointer allocated */
	size_t offset;	/* offset into current pointer (gives first available byte) */
} PointerListType, *PointerListPtr;
	
/* Allocate a string by returning the next available space in a pointer block,
	or allocate a new pointer block if the string wouldn't fit in the current
	pointer block. This assumes that strings are allocated one after the other,
	and then all strings are disposed of together. Thus, there are never any
	fragmented blocks. The returned pointer is not word alligned. The 'ptrlist'
	parameter is a linked list of all of the pointers allocated from the
	system, and is used for disposing of those pointers when finished with the
	strings. */
static char *stralloc(const char *str, size_t length, LLPtr *list)
{
	PointerListPtr *ptrlist = (PointerListPtr *) list;
	char *p;
	
	require(length + 1 < PTRSZ - sizeof(PointerListType));
	if (! *ptrlist || (*ptrlist)->offset + length + 1 >= PTRSZ) {
		*ptrlist = LLPInsert(*ptrlist, PtrBegin(PTRSZ));
		(*ptrlist)->offset = sizeof(PointerListType);
	}
	p = (char *) *ptrlist + (*ptrlist)->offset;
	(*ptrlist)->offset += length + 1;
	ensure(PtrValid(*ptrlist));
	ensure((char *) *ptrlist + sizeof(PointerListType) <= p);
	ensure(p + length + 1 < (char *) *ptrlist + PTRSZ);
	return(strcpy(p, str));
}

/* dispose of one string */
static void strfree(char *str)
{
	/* do nothing */
}

/* dispose of all strings allocated from the pointer list */
static void strfreeall(LLPtr list)
{
	PointerListPtr ptrlist = (PointerListPtr) list;

	while (ptrlist) {
		void *p = ptrlist;
		ptrlist = LLPDelete(ptrlist, ptrlist);
		PtrEnd(p);
	}
}

/*---------------------------------------------------------------------------*/
/* key comparison */
/*---------------------------------------------------------------------------*/

/* function called to compare a key while searching a tree */
static int tree_compare(const void *key, const NodePtr node)
{
	return(strcmp(key, ((NodeItemType *) node)->entry.key));
}

/* function called to compare a key while searching a hash table */
static int hash_compare(const void *key, const LLPtr item)
{
	return(strcmp(key, ((HashItemType *) item)->entry.key));
}

/* function called to two items while sorting a hash table */
static int hash_sort(const void *item1, const void *item2)
{
	return(strcmp((**(HashItemType **) item1).entry.key,
					  (**(HashItemType **) item2).entry.key));
}

/*---------------------------------------------------------------------------*/
/* table allocation */
/*---------------------------------------------------------------------------*/

/* return a new table */
LookupTableType *LkpTableBegin(LookupTableKind kind, long param)
{
	volatile LookupTableType *table = NULL;
	
	TRY {
		table = PtrBeginClear(sizeof(LookupTableType));
		FailNIL(table);
		table->kind = kind;
		switch (kind) {
		case LKP_TABLE_HASH:
			table->hash = HashBegin(param);
			break;
		case LKP_TABLE_TREE:
			table->tree = TreeBegin(param);
			break;
		case LKP_TABLE_LEXTREE:
			table->lextree = LexTreeBegin();
			break;
		}
	} CATCH {
		LkpTableEnd(table);
	} ENDTRY;
	return(table);
}

/* function called to dispose of a tree node */
static void dispose_tree(NodePtr node, void *data)
{
	NodeItemType *item = (NodeItemType *) node;
	
	if (item) {
		if (item->entry.key)
			strfree(item->entry.key);
		PtrFixedEnd(item);
	}
}

/* function called to dispose of a hash table item */
static void dispose_hash(void *p)
{
	HashItemType *item = p;
	
	if (item) {
		if (item->entry.key)
			strfree(item->entry.key);
		PtrFixedEnd(item);
	}
}

/* dispose of the table */
void LkpTableEnd(LookupTableType *table)
{
	if (table) {
		TreeEnd(table->tree, dispose_tree);
		HashEnd(table->hash, dispose_hash);
		LexTreeEnd(table->lextree);
		strfreeall(table->ptrlist);
		PtrEnd(table);
	}
}

/*---------------------------------------------------------------------------*/
/* table properties and attributes */
/*---------------------------------------------------------------------------*/

/* return number of items in table */
size_t LkpTableSize(LookupTableType *table)
{
	size_t result = 0;
	
	switch (table->kind) {
	case LKP_TABLE_HASH:
		result = HashNItems(table->hash);
		break;
	case LKP_TABLE_TREE:
		result = TreeSize(table->tree);
		break;
	case LKP_TABLE_LEXTREE:
		result = LexTreeSize(table->lextree);
		break;
	}
	return(result);
}

/*---------------------------------------------------------------------------*/
/* table insertion */
/*---------------------------------------------------------------------------*/

/* find key in table, increment counter and insert if new key */
void LkpTableInsert(LookupTableType *table,
	const void *key, size_t size)
{
	require(strlen(key) == size);
	switch (table->kind) {
	case LKP_TABLE_HASH: {
			LLPtr item;
			size_t index;
			index = HashIndex(table->hash, key, size);
			item = HashFind(table->hash, key, index, hash_compare);
			if (! item) {
				item = PtrFixedBegin(sizeof(HashItemType));
				memclr(item, sizeof(HashItemType));
				HashInsert(table->hash, item, index);
				((HashItemType *) item)->entry.key =
					stralloc(key, size, &table->ptrlist);
			}
			((HashItemType *) item)->entry.count++;
		}
		break;
	case LKP_TABLE_TREE: {
			NodePtr parent;
			NodePtr node;
			int cmp;
			node = TreeFindAt(table->tree, key, tree_compare, &parent, &cmp);
			if (! node) {
				node = PtrFixedBegin(sizeof(NodeItemType));
				memclr(node, sizeof(NodeItemType));
				TreeInsertAt(table->tree, node, parent, cmp);
				((NodeItemType *) node)->entry.key =
					stralloc(key, size, &table->ptrlist);
			}
			((NodeItemType *) node)->entry.count++;
		}
		break;
	case LKP_TABLE_LEXTREE: {
			LexNodePtr node = LexTreeFind(table->lextree, key);
			check(node->terminal);
			node->data = (void *) ((long) node->data + 1); /* use the pointer as a counter */
		}
		break;
	}
}

/*---------------------------------------------------------------------------*/
/* table output */
/*---------------------------------------------------------------------------*/

/* write key and count to file */
static void LkpTableWriteKey(LookupTableType *table, long count, const char *key)
{
	char s[MAXLEN + 32];
	int len = 0;
	
	len = sprintf(s, "%5ld\t'%s'\r", count, key);
	if (len == -1) FailOSErr(-1);
	check(0 < len && len < sizeof(s));
	FileWrite(table->fp, len, s);
	ProgressRun(table->progress, -1);
}

/* write item to file */
static void LkpTableWriteEntry(LookupTableType *table, const LkpEntryType *item)
{
	LkpTableWriteKey(table, item->count, item->key);
}

/* function called while traversing tree to write each node */
static void tree_visit(NodePtr node, void *table)
{
	LkpTableWriteEntry(table, &((NodeItemType *) node)->entry);
}

/* function called while traversing a lexical tree to write each terminal node */
static void lextree_visit(LexNodePtr node, const char *prefix,
	size_t length, void *data)
{
	LkpTableWriteKey(data, (long) node->data, prefix);
}

/* write table to file in sorted order */
void LkpTableWrite(LookupTableType *table, FileType *fp, ProgressHandle progress)
{
	CStr31 prompt;
	
	ResStrLen(RLS_BUSY, RLS_BUSY_WRITE, prompt, sizeof(prompt));
	table->progress = progress;
	table->fp = fp;
	switch (table->kind) {
	case LKP_TABLE_HASH: {
			LLPtr *sorted = NULL;
			size_t i = 0, nitems = 0;
			ProgressPrompt(progress, "Sorting words.");
			sorted = HashSort(table->hash, hash_sort);
			ProgressPrompt(progress, prompt, FileName(fp));
			nitems = HashNItems(table->hash);
			for (i = 0; i < nitems; i++)
				LkpTableWriteEntry(table, &((HashItemType *) (sorted[i]))->entry);
			PtrEnd(sorted);
		}
		break;
	case LKP_TABLE_TREE:
		ProgressPrompt(progress, prompt, FileName(fp));
		if (TreeRoot(table->tree))
			TreeInOrder(TreeRoot(table->tree), tree_visit, table);
		break;
	case LKP_TABLE_LEXTREE: {
			char prefix[MAXLEN];
			ProgressPrompt(progress, prompt, FileName(fp));
			if (LexTreeRoot(table->lextree))
				LexTreeInOrder(LexTreeRoot(table->lextree), prefix, 0, lextree_visit, table);
		}
		break;
	}
}
