
/*
	Get a list of words in a text file. Demonstrates use of progress
	dialog, folder selection and scanning, automatic opening of the
	output file, as well as the use of different data structures to
	build the list of words (binary tree, splay tree, lexical search
	tree, and hash table). It also demonstrates retrying with a
	different algorithm if the first algorithm fails.
	
	94/01/20 aih - improved testing function and added memory usage estimate
	94/01/17 aih - added retry with more memory efficient algorithm
	93/12/19 aih - split table functions into separate file
	93/10/25 aih - created
*/

#include <ctype.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "pstr.h"
#include "ApplicationLib.h"
#include "DebugLib.h"
#include "DialogLib.h"
#include "DocumentLib.h"
#include "FileBufferLib.h"
#include "FileDialogLib.h"
#include "FolderLib.h"
#include "LookupTableLib.h"
#include "LowMemLib.h"
#include "MacLib.h"
#include "MemoryLib.h"
#include "ProgressLib.h"
#include "ResourceConstantsLib.h"
#include "ResourceLib.h"
#include "WindowMenuLib.h"
#include "WordsInFile.h"

/* Define this as 1 to run tests on each table algorithm. */
#define TEST_WORDS (0)

/* Define this as TRY_LEXTREE to first try using a lexical search tree.
	If there isn't enough memory it will then retry with a hash table. */
#define TRY_FIRST	TRY_HASH

#if defined(THINK_C) && PROFILE

	/* headers needed for profiling */
	#include <profile.h>
	#include <console.h>
	#include <stdio.h>
	
	/* We don't need to profile the progress library, which we already
		know is slow (since it calls EventAvail/GNE/WNE). */
	#define ProgressRun(progress, count) ((void) 0)
#endif

/* algorithms to try */
typedef enum {
	TRY_LEXTREE,
	TRY_HASH,
	TRY_LAST
} WordsTryType;

/*---------------------------------------------------------------------------*/
/* declarations and definitions */
/*---------------------------------------------------------------------------*/

#define SUFFIX " words"	/* suffix appended to source file name to yield
									default destination file name */

/* progress dialog */
static ProgressHandle gProgress;

/* set to true to automatically open output file in a new window */
static Boolean gOpenOutput = true;

/*---------------------------------------------------------------------------*/
/* word input */
/*---------------------------------------------------------------------------*/

/* read a word with at most n-1 characters from the buffered file,
	return length of the word and terminate it with a null character */
static size_t getword(FileBufferType *buf, char *word, register short n)
{
	register char c, *p = word;

	require(n >= 2);
	while (! isalpha(c = FBufReadChar(buf)) && c != '_') {
		if (FBufEof(buf))
			return;
	}
	do {
		*p++ = c;
	} while (--n > 1 && (isalpha(c = FBufReadChar(buf)) || c == '_'));
	*p = 0;
	return(p - word);
}

/* read each word in the file and insert it into the table */
static void getwords(FileType *fp, LookupTableType *table)
{	
	FileBufferType buffer, *buf = &buffer;
	char word[MAXLEN];
	CStr31 prompt;
	size_t len;
	
	TRY {
	
		/* open file and set up progress dialog */
		FileOpen(fp, fsRdPerm);
		ResStrLen(RLS_BUSY, RLS_BUSY_READ, prompt, sizeof(prompt));
		ProgressReset(gProgress, FileSize(fp));
		ProgressPrompt(gProgress, prompt, FileName(fp));
				
		/* read each word from file and add to table if it's a new word,
			otherwise increment counter for the word */
		FBufInit(buf, fp);
		do {
			len = getword(buf, word, MAXLEN);
			if (len > 0 && table)
				LkpTableInsert(table, word, len);
			ProgressRun(gProgress, buf->position - (buf->size - buf->index));
		} while (! FBufEof(buf));
		ProgressUpdate(gProgress);
	} CLEANUP {
		FileClose(fp);
	} ENDTRY;
}

/*---------------------------------------------------------------------------*/
/* word output */
/*---------------------------------------------------------------------------*/

/* write words in sorted order to the file */
static void printwords(FileType *fp, LookupTableType *table)
{
	TRY {
		ProgressReset(gProgress, LkpTableSize(table));
		if (FileExists(fp))
			FileDelete(fp);
		FileCreate(fp, AppCreator(), 'TEXT');
		FileOpen(fp, fsWrPerm);
		LkpTableWrite(table, fp, gProgress);
		ProgressUpdate(gProgress);
		FileClose(fp);
	} CATCH {
		FileClose(fp);
	} ENDTRY;
}

/* open file containing words in a new window; 'suffix' is appended to the
	title of the window */
static void openwords(FileType *fp, const char *suffix)
{
	WindowPtr window;		/* opened output file's window */
	CStr255 title;			/* title for window */

	if (gOpenOutput) {
		window = DocWindow(DocBegin(fp));
		if (suffix) {
			WinMenuRemove(window);
			WinTitle(window, title);
			WinTitleSet(window, strcat(strcat(title, " - "), suffix));
			WinMenuAdd(window);
		}
		WinShow(window);
	}
}

/*---------------------------------------------------------------------------*/
/* scanning */
/*---------------------------------------------------------------------------*/

/* function called for each file and folder while scanning folders */
static Boolean action(const CInfoPBRec *pb, void *table)
{
	FileType file, *fp = &file;
	CStr255 name;
	
	/* process only files of type 'TEXT' */
	if ((pb->hFileInfo.ioFlAttrib & (1<<kFolderBit)) == 0 &&
		  pb->hFileInfo.ioFlFndrInfo.fdType == 'TEXT')
	{
		/* process words in file */
		p2cstrcpy(name, pb->hFileInfo.ioNamePtr);
		FileSet(fp, pb->hFileInfo.ioVRefNum, pb->hFileInfo.ioFlParID, name);
		getwords(fp, table);
	}
}

/* scan a file or folder */
static void scanwords(FileType *fp, LookupTableType *table)
{
	if (FolderExists(fp))
		FolderScan(fp, true, action, table);
	else
		getwords(fp, table);
}

/* list words in source file, write results to destination file */
static void DoWordsInFile(FileType *src, FileType *dst)
{
	volatile LookupTableType *table = NULL;/* table containing words */
	static WordsTryType try = TRY_FIRST;	/* first algorithm to try */
	
	TRY {
		/* A lexical search tree is the fastest alogrithm, followed by
			a hash table. However, lexical search trees typically use
			much more memory than a hash table. We first try using a lexical
			search tree, but if we run out of memory we switch to a hash
			table. To keep from having to rediscover that we're low on
			memory each time we're executed, the flag that determines which
			algorithm is used is a static variable. */
		switch (try) {
		case TRY_LEXTREE:
			/* try with a lexical search tree */
			table = LkpTableBegin(LKP_TABLE_LEXTREE, 0);
			scanwords(src, table);
			printwords(dst, table);
			break;
		default:
			/* lexical search tree failed, so try a hash table */
			table = LkpTableBegin(LKP_TABLE_HASH, HashCalculateSize(HASH_MINSIZE * 4, 10 * CLOCKS_PER_SEC));
			scanwords(src, table);
			printwords(dst, table);
			break;
		}
		LkpTableEnd(table);
		table = NULL;
		openwords(dst, NULL);
	} CATCH {
		LkpTableEnd(table);
		table = NULL;
		if (try < TRY_HASH && FailReason() == memFullErr) {
			/* ran out of memory, retry with a more memory efficient algorithm */
			try++;
			RETRY;
		}
	} ENDTRY;
}

#if TEST_WORDS

#if defined(THINK_C) && PROFILE
	#define DebugPrintf printf
	console_options.nrows = 10;
	cecho2file("console", 0, stdout);
	#define PROFILE_START	{ResetProfile(); _profile = true; }
	#define PROFILE_STOP		{_profile = false; DumpProfile(); }
#else
	#define PROFILE_START
	#define PROFILE_STOP
#endif		

/* run a test on a lookup table */
static void DoTestWordsInFile(FileType *src, FileType *dst,
	LookupTableKind kind, long param, const char *prompt)
{
	volatile LookupTableType *table = NULL;/* table containing words */
	long membefore, memafter; /* for estimating memory usage */

	TRY {
		DebugPrintf("\nTesting a %s:\n", prompt);
		PROFILE_START;
		table = LkpTableBegin(kind, param);
		scanwords(src, table);
		printwords(dst, table);
		DebugPrintf("Number of unique words = %lu\n", LkpTableSize(table));
		membefore = FreeMem();
		LkpTableEnd(table);
		table = NULL;
		memafter = FreeMem();
		DebugPrintf("Memory used by table = %ld bytes\n", memafter - membefore);
		PROFILE_STOP;
		openwords(dst, prompt);
	} CATCH {
		LkpTableEnd(table);
	} ENDTRY;
}

/* run tests on each lookup table algorithm */
static void TestWordsInFile(FileType *src, FileType *dst)
{
	static struct { /* static so can easily change parameters in the debugger */
		Boolean assertions:1;
		Boolean cache:1;
		Boolean tree:1;
		Boolean splay:1;
		Boolean lextree:1;
		Boolean hash:1;
	} test = { ! PROFILE, true, true, true, true, true, };
	struct { /* saved assertion flags */
		Boolean require:1;
		Boolean ensure:1;
		Boolean check:1;
	} assert;
	
	TRY {
	
		/* setup */
		assert.require = _assert.require;
		assert.ensure = _assert.ensure;
		assert.check = _assert.check;
		_assert.require = _assert.ensure = _assert.check = test.assertions;
		
		/* print settings affecting all tests */
		DebugPrintf("assertions = %s, profile = %s\n",
			test.assertions ? "true" : "false",
			PROFILE ? "true" : "false");
		
		/* run tests */
		if (test.cache) {
			DebugPrintf("\nCaching file\n");
			PROFILE_START;
			scanwords(src, NULL);
			PROFILE_STOP;
		}
		if (test.tree) {
			DoTestWordsInFile(src, dst, LKP_TABLE_TREE, false,
				"binary tree");
		}		
		if (test.splay) {
			DoTestWordsInFile(src, dst, LKP_TABLE_TREE, true,
				"splay tree");
		}
		if (test.lextree) {
			DoTestWordsInFile(src, dst, LKP_TABLE_LEXTREE, 0,
				"lexical search tree");
		}
		if (test.hash) {
			DoTestWordsInFile(src, dst,
				LKP_TABLE_HASH, HashCalculateSize(HASH_MINSIZE * 4, 10 * CLOCKS_PER_SEC),
				"hash table");
		}
	} CLEANUP {
		_assert.require = assert.require;
		_assert.ensure = assert.ensure;
		_assert.check = assert.check;
	} ENDTRY;
}

#endif /* TEST_WORDS */

/* build a list of the words in a file or folder */
void WordsInFile(FileType *src, FileType *dst)
{
	FileType srcfile;					/* source file if src is NULL */
	FileType dstfile;					/* destination file if dst is NULL */
	FileNameType name;				/* default file name for file dialog */
	CStr31 prompt;						/* prompt for file dialog */
	OSType textType = 'TEXT';		/* for selecting text files */
	
	TRY {
	
		/* select source file or folder */
		if (! src) {
			src = &srcfile;
			if (MacHasStandardFile())
				FileOrFolderSelect(src, 1, &textType);
			else
				FileSFGet(src, 1, &textType);
		}
		
		/* select destination */
		if (! dst) {
			dst = &dstfile;
			/* append suffix to source file name and use as default file name */
			strcpy(name, FileName(src));
			name[sizeof(FileNameType) - strlen(SUFFIX) - 1] = 0;
			strcat(name, SUFFIX);
			FileSet(dst, -GetSFSaveDisk(), GetCurDirStore(), name);
			FileUnique(dst);
			strcpy(name, FileName(dst));
			ResStrLen(RLS_SF, RLS_SF_SAVE, prompt, sizeof(prompt));
			FileSFPut(dst, prompt, name);
		}
		
		/* create progress dialog */
		check(! gProgress);
		gProgress = ProgressBegin(0, NULL, NULL, 0, 0);
		
		#if ! TEST_WORDS
			DoWordsInFile(src, dst);
		#else
			TestWordsInFile(src, dst);
		#endif
		
	} CLEANUP {
		ProgressEnd(gProgress);
		gProgress = NULL;
	} ENDTRY;
}
